# Maxime MORGE <maxime.morge@univ-lyon1.fr>
import os
import asyncio
import json
from typing import Literal, Dict
from pydantic import BaseModel
from autogen_agentchat.agents import AssistantAgent
from autogen_agentchat.messages import TextMessage
from autogen_core import CancellationToken
from autogen_ext.models.openai import OpenAIChatCompletionClient
from welfare import Welfare

# Load API key from environment variable
OPENAI_API_KEY = os.getenv("OPENAI_API_KEY")
# Validate API key
if not OPENAI_API_KEY:
    raise ValueError("Missing OPENAI_API_KEY. Set it as an environment variable.")


class AgentResponse(BaseModel):
    my_share: Literal[500, 100, 400, 325]
    other_share: Literal[100, 500, 300, 325]
    lost: Literal[400, 400, 300, 350]
    motivation: str

    @classmethod
    def validate_combination(cls, values):
        """Ensure (my_share, other_share, lost) is a valid predefined combination"""
        valid_combinations = {
            (500, 100, 400),
            (100, 500, 400),
            (400, 300, 300),
            (325, 325, 350)
        }
        if (values.get("my_share"), values.get("other_share"), values.get("lost")) not in valid_combinations:
            raise ValueError("Invalid share combination")
        return values

    class Config:
        validate_assignment = True  # Ensure values are always validated


class DictatorSetup:
    debug = False

    def __init__(self, model: str, temperature: float, preference: Welfare, strategy=False):
        self.model = model
        self.temperature = temperature
        self.preference = preference
        self.strategy = strategy

        if not self.strategy:
            is_openai_model = model.startswith("gpt")
            base_url = "https://api.openai.com/v1" if is_openai_model else "http://localhost:11434/v1"

            self.model_info = {
                "temperature": self.temperature,
                "function_calling": True,
                "parallel_tool_calls": True,
                "family": "unknown",
                "json_output": True,
                "vision": False
            }

            self.model_client = OpenAIChatCompletionClient(
                model=self.model,
                base_url=base_url,
                api_key=OPENAI_API_KEY,
                model_info=self.model_info,
                response_format=AgentResponse
            )

    async def run(self) -> Dict:
        """Runs the dictator game and returns a detailed response dictionary."""
        if self.strategy:
            return self.apply_strategy(self.model)

        instruction = f"""
        Hi, let’s play a game. You are paired with another player.
        Your role is to decide how to divide $1000, and the other player simply receives your choice.
        Money can be lost in the process.
        {self.preference.description}
        
        How would you like to divide the money?
    
        Choose one of the following options:
        - my_share: 500, other_share: 100, lost: 400
        - my_share: 100, other_share: 500, lost: 400
        - my_share: 400, other_share: 300, lost: 300
        - my_share: 325, other_share: 325, lost: 350
    
        Please provide one specific choice, stating how much you keep, how much you give, and how much is lost.
        """
        agent = AssistantAgent(
            name="Dictator",
            model_client=self.model_client,
            system_message="You are a helpful assistant."
        )
        response = await agent.on_messages(
            [TextMessage(content=instruction, source="user")],
            cancellation_token=CancellationToken(),
        )

        response_data = response.chat_message.content
        if self.debug:
            print(f"Raw Response: {response_data}")

        response_dict = json.loads(response_data)
        agent_response = AgentResponse.model_validate(response_dict)

        is_consistent = self.check_consistency(agent_response)
        return {
            "is_consistent": is_consistent,
            "my_share": agent_response.my_share,
            "other_share": agent_response.other_share,
            "lost": agent_response.lost,
            "motivations": agent_response.motivation
        }

    def apply_strategy(self, model: str) -> Dict:
        """Applies a predefined strategy based on the preference."""
        if model == "gpt-4.5-preview-2025-02-27":
            strategy_mapping = {
                Welfare.SELFISH: {"my_share": 500, "other_share": 100, "lost": 400, "motivations": "Maximizing self-interest"},
                Welfare.ALTRUISTIC: {"my_share": 100, "other_share": 500, "lost": 400, "motivations": "Helping others at a personal cost"},
                Welfare.UTILITARIAN: {"my_share": 400, "other_share": 300, "lost": 300, "motivations": "Maximizing total utility"},
                Welfare.EGALITARIAN: {"my_share": 325, "other_share": 325, "lost": 350, "motivations": "Ensuring fairness"}
            }
            return strategy_mapping.get(self.preference, {"error": "Preference strategy not defined"})
        if model == "llama3":
            strategy_map = {
                Welfare.SELFISH: (500, 100, 400),
                Welfare.ALTRUISTIC: (100, 500, 400),
                Welfare.UTILITARIAN: (400, 300, 300),
                Welfare.EGALITARIAN: (325, 325, 350)
            }
            if self.preference in strategy_map:
                my_share = strategy_map[self.preference][0]
                other_share = strategy_map[self.preference][1]
                lost = strategy_map[self.preference][2]
                return {
                    "my_share": my_share,
                    "other_share": other_share,
                    "lost": lost,
                    "is_consistent": True,
                }
        else:
            raise ValueError("Invalid preference type")
            return  {"error": "Preference strategy not defined"}
        if model == "mistral-small":
            valid_choices = {
                Welfare.SELFISH: {"my_share": 500, "other_share": 100, "lost": 400},
                Welfare.ALTRUISTIC: {"my_share": 100, "other_share": 500, "lost": 400},
                Welfare.UTILITARIAN: {"my_share": 400, "other_share": 300, "lost": 300},
                Welfare.EGALITARIAN: {"my_share": 325, "other_share": 325, "lost": 350},
            }
            strategy = valid_choices.get(self.preference)
            if not strategy:
                raise ValueError(f"Unknown preference type {self.preference}")
            return {
                "is_consistent": True,
                **strategy
            }
        if model == "deepseek-r1":
            return  {"error": "Preference strategy not defined"}


    def check_consistency(self, agent_response: AgentResponse) -> bool:
        """Check if the response aligns with the given preference."""
        valid_choices = {
            Welfare.SELFISH: (500, 100, 400),
            Welfare.ALTRUISTIC: (100, 500, 400),
            Welfare.UTILITARIAN: (400, 300, 300),
            Welfare.EGALITARIAN: (325, 325, 350),
        }

        expected_values = valid_choices.get(self.preference, None)
        if expected_values:
            return (
                    agent_response.my_share == expected_values[0] and
                    agent_response.other_share == expected_values[1] and
                    agent_response.lost == expected_values[2]
            )
        return False


# Run the async function and return the response
if __name__ == "__main__":
    preference = Welfare.EGALITARIAN
    game_agent = DictatorSetup(model="llama3", temperature=0.7, preference=preference, strategy=True)
    response = asyncio.run(game_agent.run())
    print(response)
