import os
import asyncio
import csv
import random
from typing import Dict, Literal, List, Callable
from pydantic import BaseModel, ValidationError
from autogen_agentchat.agents import AssistantAgent
from autogen_agentchat.messages import TextMessage
from autogen_core import CancellationToken
from autogen_ext.models.openai import OpenAIChatCompletionClient
import json

# Load API key from environment variable
OPENAI_API_KEY = os.getenv("OPENAI_API_KEY")
if not OPENAI_API_KEY:
    raise ValueError("Missing OPENAI_API_KEY. Set it as an environment variable.")

CSV_FILE_PATH = "../../data/rps/rps.csv"

# Define the expected response format as a Pydantic model
class AgentResponse(BaseModel):
    prediction: Literal["Rock", "Paper", "Scissor"]
    reasoning: str

# Define Guess simulation class
class Guess:
    def __init__(self, model: str, temperature: float, game_id: int, opponent_strategy_fn: Callable[[List[Dict]], str], strategy=False, max_retries: int = 3):
        self.model = model
        self.temperature = temperature
        self.game_id = game_id
        self.max_retries = max_retries
        self.history: List[Dict] = []
        self.player_score_game = 0
        self.opponent_strategy_fn = opponent_strategy_fn
        self.strategy = strategy  # Determines whether to use a model or a rule-based method

        if not strategy:  # Use model-based prediction
            is_openai_model = model.startswith("gpt")
            base_url = "https://api.openai.com/v1" if is_openai_model else "http://localhost:11434/v1"
            model_info = {
                "temperature": self.temperature,
                "function_calling": True,
                "parallel_tool_calls": True,
                "family": "unknown",
                "json_output": True,
                "vision": False
            }
            self.model_client = OpenAIChatCompletionClient(
                model=self.model,
                base_url=base_url,
                api_key=OPENAI_API_KEY,
                model_info=model_info,
                response_format=AgentResponse
            )
        else:
            self.model_client = None  # No model needed for rule-based strategy

    async def play_round(self, round_id: int) -> Dict:
        """Plays a single round of Guess The Next Move."""
        opponent_move = self.opponent_strategy_fn(self.history)

        if self.strategy:
            # Use a simple rule-based prediction
            move, reasoning = self.apply_strategy()
        else:
            # Use the model for prediction
            move, reasoning = await self.model_based_prediction()

        outcome = self.determine_accuracy(move, opponent_move)
        self.update_score(outcome)
        round_result = {
            "Agent Prediction": move,
            "Opponent Move": opponent_move,
            "Reasoning": reasoning,
            "Outcome": outcome
        }
        self.history.append(round_result)
        return round_result

    async def model_based_prediction(self):
        """Generates a move using the AI model."""
        history_summary = self.get_history_summary()
        instruction = f"""
        We are playing a simplified version of Rock-Paper-Scissors (RPS) designed to evaluate your ability to predict the opponent's next move.

        ### **Game Rules:**
        1. The opponent follows a hidden strategy (which may involve a repeating pattern or adaptive behavior).
        2. Your task is to predict the opponent’s next move: Rock, Paper, or Scissors.
        3. If your prediction is correct, you earn **1 point**; otherwise, you earn **0 points**.
        4. The game continues for multiple rounds, and your accuracy is evaluated at each round.

        ### **Game History So Far:**
        {history_summary}

        ### **Your Task:**
        Based on the game history, predict the opponent's next move.  
        Return your response in JSON format with two keys:  
        - `"prediction"`: Your predicted move (`"Rock"`, `"Paper"`, or `"Scissor"`).  
        - `"reasoning"`: A brief explanation of how you made your prediction.
        """
        for attempt in range(1, self.max_retries + 1):
            agent = AssistantAgent(
                name="Player",
                model_client=self.model_client,
                system_message="You are a helpful assistant."
            )
            response = await agent.on_messages(
                [TextMessage(content=instruction, source="user")],
                cancellation_token=CancellationToken(),
            )
            try:
                response_data = response.chat_message.content
                agent_response = AgentResponse.model_validate_json(response_data)
                move, reasoning = agent_response.prediction, agent_response.reasoning

                if move in ["Rock", "Paper", "Scissor"]:
                    return move, reasoning
            except (ValidationError, json.JSONDecodeError) as e:
                print(f"Error parsing response (Attempt {attempt}): {e}")
        raise ValueError("Model failed to provide a valid response after multiple attempts.")

    def apply_strategy(self):
        """Predicts the next move using a heuristic."""
        if self.model == "gpt-4.5-preview-2025-02-27":
            if not self.history:
                return random.choice(["Rock", "Paper", "Scissor"]), "No history available. Choosing randomly."
            # Count occurrences of each move
            move_counts = {"Rock": 0, "Paper": 0, "Scissor": 0}
            for round_data in self.history:
                move_counts[round_data["Opponent Move"]] += 1
            # Find the most common move
            most_common_move = max(move_counts, key=move_counts.get)
            predicted_move = most_common_move
            reasoning = f"Based on history, the opponent most frequently played {most_common_move}."
            return predicted_move, reasoning
        if self.model == "llama3":
             return ["None", "error"]
        if self.model == "mistral-small":
            if not self.history:
                # If there is no history, we can't make an educated guess.
                return ["Scissor", "No game history available."]
            opponent_moves = [move['Opponent Move'] for move in self.history]
            move_count = {
                'Rock': opponent_moves.count('Rock'),
                'Paper': opponent_moves.count('Paper'),
                'Scissors': opponent_moves.count('Scissor')
            }
            # Determine the most frequent move
            max_move = max(move_count, key=move_count.get)
            if move_count[max_move] > 0:
                reasoning = f"Predicted {max_move} because it has been played {move_count[max_move]} times."
            else:
                reasoning = "Unable to determine a pattern; defaulting to Scissors."
            return max_move, reasoning
        if self.model == "deepseek-r1":
            return ["None", "error"]

    @staticmethod
    def determine_accuracy(player_move: str, opponent_move: str) -> int:
        """Determines the accuracy of the prediction."""
        return 1 if player_move == opponent_move else 0

    def update_score(self, outcome: int):
        """Updates the score based on the outcome."""
        if outcome == 1:
            self.player_score_game += 1

    def get_history_summary(self) -> str:
        """Summarizes the game history for model-based predictions."""
        if not self.history:
            return "This is the first round."
        summary = "\n".join(
            [f"Round {i+1}: You guessed {r['Agent Prediction']}, Opponent played {r['Opponent Move']}. Outcome: {r['Outcome']}"
             for i, r in enumerate(self.history)]
        )
        summary += f"\nCurrent Score - You: {self.player_score_game}\n"
        return summary

def simple_opponent_strategy(history):
    """A simple opponent strategy that cycles through Rock, Paper, Scissor."""
    moves = ["Rock", "Paper", "Scissor"]
    return moves[len(history) % 3]

async def main():
    # Play with strategy-based approach
    game = Guess(model="mistral-small", temperature=0.7, game_id=1, opponent_strategy_fn=lambda history: "Rock", strategy=True)
    num_rounds = 10
    for round_id in range(1, num_rounds + 1):
        result = await game.play_round(round_id)
        print(f"Round {round_id}: {result}")
    print(f"Final Score: {game.player_score_game}")

if __name__ == "__main__":
    asyncio.run(main())